/*   This file is part of Armadillo Music Player.
*
*   Armadillo is free software: you can redistribute it and/or modify
*   it under the terms of the GNU General Public License as published by
*   the Free Software Foundation, either version 3 of the License, or
*   (at your option) any later version.
*
*   Armadillo is distributed in the hope that it will be useful,
*   but WITHOUT ANY WARRANTY; without even the implied warranty of
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*   GNU General Public License for more details.
*
*   You should have received a copy of the GNU General Public License
*   along with Armadillo.  If not, see <http://www.gnu.org/licenses/>.
*/

/**
    Contains methods to operate on the SQLite database used by the Armadillo Music Player.
    Author: Katarzyna Zien for the Armadillo team :-)
 **/

package armadilloUI;

import java.io.File;
import java.sql.*;
import java.util.*;
import java.util.logging.*;
/**
 * ArmadilloDB class, provides all the method and query to access data within the
 * database, also deals with multiple threads accessing it.
 * @author kasia
 */
public class ArmadilloDB extends DbConnection
{
    private static String[] tables = {"track", "tagtype", "tag","moodmodel","playlist","track_tags", "playlist_track", "fingerprint", "fingerprintDoubles", "fingerprintInts", "importQueue", "navTree"};
    static boolean testing = false; // CHANGE TO FALSE WHEN NOT TESTING

    /**
     * Creates a DB object and initialises it, ready to query.
     * @param sDriverToLoad
     * @param sUrlToLoad
     * @throws Exception
     */
    public ArmadilloDB(String sDriverToLoad, String sUrlToLoad) throws Exception
    {
        ArmadilloLog.LogOperation("Constructing ArmadilloDB object");
        init(sDriverToLoad, sUrlToLoad);
    }

    /** Create tables needed: track, tagtype (tagtype_id, type, sort), tag, moodmodel, playlist, track_tags, playlist_track. *
     * @throws Exception
     */
    public void createTables() throws Exception
    {
        for (int i=0; i<tables.length; i++)
        {
          this.executeStmt("drop table if exists "+tables[i]+";");
        }
        
        // table TRACK: track_id, location, duration, analysed
        this.executeStmt("CREATE TABLE track (track_id INTEGER PRIMARY KEY, location TEXT NOT NULL, duration LONG NOT NULL, analysed TEXT NOT NULL);");
        // table TAGTYPE: tagtype_id, type
        this.executeStmt("CREATE TABLE tagtype (tagtype_id INTEGER PRIMARY KEY, type TEXT NOT NULL);");
        this.executeStmt("CREATE INDEX tag_type_index ON tagtype (type);");
            //create default tag types: TITLE, ARTIST, ALBUM, GENRE, BPM
            insertTagType("Title");
            insertTagType("Artist");
            insertTagType("Album");
            insertTagType("Genre");
            insertTagType("BPM");
        // table TAG: tag_id, value, tagtype_id
        this.executeStmt("CREATE TABLE tag (tag_id INTEGER PRIMARY KEY, value TEXT NOT NULL, tagtype_id INTEGER NOT NULL);");
            //create default 'Unknown' tags
            insertTag("<Unknown>", 2);
            insertTag("<Unknown>", 3);
            insertTag("<Unknown>", 4);
        this.executeStmt("CREATE INDEX value_index ON tag (value);");
        this.executeStmt("CREATE INDEX tagtype_id_index ON tag (tagtype_id);");
        // table PLAYLIST: playlist_id, name, now_playing
        this.executeStmt("CREATE TABLE playlist (playlist_id INTEGER PRIMARY KEY, name TEXT NOT NULL, now_playing INTEGER NOT NULL);");
            //create default playlists
            insertPlaylist("Last played", 1);
            insertPlaylist("Last searched", 1);
        this.executeStmt("CREATE INDEX name_index ON playlist (name);");
        // table TRACK_TAGS: tag_id, track_id
        this.executeStmt("CREATE TABLE track_tags (tag_id INTEGER NOT NULL, track_id INTEGER NOT NULL);");
        // table PLAYLIST_TRACK: playlist_id, track_id, position
        this.executeStmt("CREATE TABLE playlist_track (playlist_id INTEGER NOT NULL, track_id INTEGER NOT NULL, position INTEGER NOT NULL);");
        this.executeStmt("CREATE INDEX playlist_id_index ON playlist_track (playlist_id);");
        // table FINGERPRINT: track_id, averageDec, ...
        this.executeStmt("CREATE TABLE fingerprint (track_id INTEGER NOT NULL, averageDec DOUBLE);");
        // table FINGERPRINTDOUBLES: track_id, 1..49
        this.executeStmt("CREATE TABLE fingerprintDoubles (track_id INTEGER NOT NULL);");
        // table FINGERPRINTINTS: track_id, 1..49
        this.executeStmt("CREATE TABLE fingerprintInts (track_id INTEGER NOT NULL);");
        this.executeStmt("CREATE INDEX FPDoubles_index ON fingerprintDoubles (track_id);");
        this.executeStmt("CREATE INDEX FPInts_index ON fingerprintInts (track_id);");
        for (int i=1; i<51; i++)
            this.executeStmt("ALTER TABLE fingerprintDoubles ADD COLUMN '"+i+"' DOUBLE;");
        for (int j=1; j<51; j++)
            this.executeStmt("ALTER TABLE fingerprintInts ADD COLUMN '"+j+"' INTEGER;");
        this.executeStmt("CREATE TABLE importQueue (batch_no INTEGER NOT NULL, track_id INTEGER NOT NULL);");
        //// table NAVTREE: artist (NAME!), tag_id ( = album_id associated with it)
        this.executeStmt("CREATE TABLE navTree (artist TEXT NOT NULL, tag_id INTEGER NOT NULL);");

    }

    /** Store new track in DB. Return ID of the new track *
     * @param path
     * @param length
     * @param metadata
     * @return assigned track id for passed track
     */
    @SuppressWarnings("empty-statement")
    public int insertTrack(String path, long length, String[] metadata)
    {
        int newTrackID = -1;
        int albumID = -1;

        try
        {
            // add track to 'track' table
            PreparedStatement prepStmt = this.conn.prepareStatement("INSERT INTO track VALUES (NULL, ?, ?, ?);");
            prepStmt.setString(1, path);
            prepStmt.setLong(2, length);
            prepStmt.setString(3, "false");
            prepStmt.addBatch();
            prepStmt.executeBatch();
            ResultSet rs = prepStmt.getGeneratedKeys();
            while (rs.next())
                newTrackID  = rs.getInt(1);
            rs.close();

            // check if tags already exist and if not
            // create new tags representing Album, Artist, Genre
            int id;
            // CHECK TITLE, ARTIST, ALBUM, GENRE TAG
            for (int i=0; i<4; i++)
            {
                id = getTag(metadata[i], i+1);
                if ( id == -1 & newTrackID != -1)
                {
                    id = insertTag(metadata[i], i+1);
                    tagTrack (id, newTrackID);
                }
                else if ( id > -1 & newTrackID != -1 )
                {
                    tagTrack (id, newTrackID);
                }
                else
                    ArmadilloLog.LogError("ArmadilloDB could not tag track");
                if (i==2)
                    albumID = id;                
            }
            // associate album with artist
            String artist = metadata[1];
            if ( !(checkAlbum(artist, albumID)) )
                addAlbumToArtist(artist, albumID);
            return newTrackID;
        }
        catch (SQLException e)
        {
            ArmadilloLog.LogError("ArmadilloDB SQL Error - "+e.toString());
            return -1;
        }
    }

    /** Change the flag to true if the bpm and mood of the song've been analysed. *
     * @param value
     * @param trackID
     */
    public void setAnalysed(String value, int trackID)
    {
        String command = "UPDATE 'track' SET analysed = '"+ value +"' WHERE track_id = " + trackID + ";";
        try {
            this.executeStmt(command);
            
        } catch (SQLException ex) {
            ArmadilloLog.LogError("ArmadilloDB Set-Analysed Error - "+ex.toString());
        }
    }

    /** Check if the song has been analysed. *
     * @param trackID
     * @return analysed?
     */
    @SuppressWarnings("empty-statement")
    public boolean getAnalysed(int trackID)
    {
        try
        {
            ResultSet list = this.executeQry("SELECT analysed FROM 'track' WHERE track_id = " + trackID + ";");
            while (list.next())
            {
                if (list.getString("analysed").equals("true"))
                    return true;
            }
            list.close();
        }
        catch (SQLException e)
        {
            ArmadilloLog.LogError("ArmadilloDB SQL Error - "+e.toString());
        }
        return false;
    }

    /** Check if the album has already been associated with this artist. *
     * @param artist
     * @param id
     * @return album is associated
     */
    public boolean checkAlbum(String artist, int id)
    {
        try {
            artist = artist.replace("'", "''");
            ResultSet list = this.executeQry("SELECT * FROM 'navTree' WHERE artist = '" + artist + "' AND tag_id = " + id + " ;");
            while (list.next()) {
                return true;
            }
        } catch (SQLException ex) {
            Logger.getLogger(ArmadilloDB.class.getName()).log(Level.SEVERE, null, ex);
        }
        return false;
    }

    /** Associate this album with the given artist. *
     * @param artist
     * @param albumID
     */
    public void addAlbumToArtist(String artist, int albumID)
    {
        try
        {
            PreparedStatement prepStmt = conn.prepareStatement("INSERT INTO navTree VALUES (?, ?);");
            prepStmt.setString(1, artist);
            prepStmt.setInt(2, albumID);
            prepStmt.addBatch();
            prepStmt.executeBatch();
        }
        catch (SQLException e)
        {
            System.err.println(e);
        }
    }

    /**
     * Takes in an artist name and passed back all that artists album tags
     * used primarily in the nav tree
     * @param artist
     * @return all albums associated with given artist
     */
    public ArrayList<String[]> getAlbums(String artist)
    {
        ArrayList<String[]> list = new ArrayList<String[]>();
        try {
            String[] albumData = new String[2];
            artist = artist.replace("'", "''");
            ResultSet set = this.executeQry("SELECT tag_id, value FROM 'navTree' NATURAL JOIN 'tag' WHERE artist = '" + artist + "' ;");
            while (set.next()) {
                albumData[0] = set.getString("value"); // value
                albumData[1] = set.getInt("tag_id") + ""; //id
                list.add(albumData);
            }
            set.close();
        } catch (SQLException ex) {
            Logger.getLogger(ArmadilloDB.class.getName()).log(Level.SEVERE, null, ex);
        }
        return list;
    }

    /** Update location of a track. *
     * @param trackID
     * @param updatedFile
     */
    public void updateLocation(int trackID, File updatedFile)
    {
        String command = "UPDATE 'track' SET location = '"+ updatedFile.getPath() +"' WHERE track_id = " + trackID + ";";
        try {
            this.executeStmt(command);

        } catch (SQLException ex) {
            //System.out.println(ex.getStackTrace());
            ArmadilloLog.LogError("ArmadilloDB SQL Error, updating location - "+ex.toString());
        }
    }
    
    /** Store new tag type in DB. *
     * @param type
     */
    public void insertTagType (String type)
    {
        try
        {
            PreparedStatement prepStmt = conn.prepareStatement("INSERT INTO tagtype VALUES (NULL, ?);");
            prepStmt.setString(1, type);
            prepStmt.addBatch();
            prepStmt.executeBatch();
        }
        catch (SQLException e)
        {
            ArmadilloLog.LogError("ArmadilloDB SQL Error - "+e.toString());
        }
    }

    /** Store new tag in DB. *
     * @param value
     * @param tagtypeID
     * @return the new tagID associated with the created tag
     */
    public int insertTag(String value, int tagtypeID)
    {
        int newTagID = -1;
        try
        {
            PreparedStatement prepStmt = conn.prepareStatement("INSERT INTO tag VALUES (NULL, ?, ?);");
            prepStmt.setString(1, value);
            prepStmt.setInt(2, tagtypeID);
            prepStmt.addBatch();
            prepStmt.executeBatch();
            ResultSet rs = prepStmt.getGeneratedKeys();
            while (rs.next())
                newTagID  = rs.getInt(1);
            rs.close();
        }
        catch (SQLException e)
        {
            System.err.println(e);
        }
        return newTagID;
    } 

    /** Store new, empty playlist in DB. *
     * @param name
     * @param now_playing
     */
    public void insertPlaylist(String name, int now_playing)
    {
        try
        {
            PreparedStatement prepStmt = conn.prepareStatement("INSERT INTO playlist VALUES (NULL, ?, ?);");
            prepStmt.setString(1, name);
            prepStmt.setInt(2, now_playing);
            prepStmt.addBatch();
            prepStmt.executeBatch();
        }
        catch (SQLException e)
        {
            System.err.println(e);
        }
    }

    /** Tag (ID) a track (ID). *
     * @param tagID
     * @param trackID
     */
    public void tagTrack(int tagID, int trackID)
    {
        try
        {
            PreparedStatement prepStmt = conn.prepareStatement("INSERT INTO track_tags VALUES (?, ?);");
            prepStmt.setInt(1, tagID);
            prepStmt.setInt(2, trackID);
            prepStmt.addBatch();
            prepStmt.executeBatch();
        }
        catch (SQLException e)
        {
            System.err.println(e);
        }
    }

    /** Returns playlist ID given its name. Returns first occurence of a playlist in DB. *
     * @param name
     * @return playlist Id associated with the given playlist name
     */
    // RETURN INDEX OF THE NEW PLAYLIST
    public int getPlayListID(String name)
    {
        int index = -1;
        try
        {
            ResultSet list = this.executeQry("SELECT * FROM 'playlist' WHERE name = \""+name+"\";");
            while (list.next())
                index  = list.getInt("playlist_id");
            list.close();
        }
        catch (SQLException e)
        {
            System.err.println(e);
        }
        return index;
    }

    /** Returns playlist name given its ID. *
     * @param ID
     * @return playlist name associated with given playlist ID
     */
    public String getPlayListName(int ID)
    {
        String name = "<Unknown>";
        try
        {
            ResultSet list = this.executeQry("SELECT * FROM 'playlist' WHERE playlist_id = "+ID+";");
            while (list.next())
                name  = list.getString("name");
            list.close();
        }
        catch (SQLException e)
        {
            System.err.println(e);
        }
        return name;
    }

    /** Returns names of all playlists stored in DB. *
     * @return  all playlist names
     * @throws Exception
     */
    public ArrayList<String> getPlaylistsNames() throws Exception
    {
      ArrayList<String> playlists = new ArrayList<String>();
      ResultSet list = this.executeQry("SELECT * FROM 'playlist';");
      while (list.next())
          playlists.add(list.getString("name"));
      list.close();
      return playlists;
    }

    /** Returns names and IDs of all playlists stored in DB. *
     * @return all playlist ids
     * @throws Exception
     */
    public ArrayList<String[]> getPlaylists() throws Exception
    {
      ArrayList<String[]> playlists = new ArrayList<String[]>();
      ResultSet list = this.executeQry("SELECT * FROM 'playlist';");
      while (list.next())
      {
          String[] playlist = new String[2];
          playlist[0] = list.getString("name");
          playlist[1] = list.getInt("playlist_id") + "";
          playlists.add(playlist);
      }
      list.close();
      return playlists;
    }

    /** 
     * Returns an ArrayList< String[ ] > of all tags of the given tag type (eg all Artists).
     * [0] - value
     * [1] - tag id
     *
     * @param tagTypeID
     * @return all tags of given type
     */
    public ArrayList<String[]> getTags(int tagTypeID)
    {
        ArrayList<String[]> tags = new ArrayList<String[]>();
        try {
            ResultSet list = this.executeQry("SELECT * FROM 'tag' WHERE tagtype_id = " + tagTypeID + " ORDER BY value ;");
            while (list.next())
            {
                String[] properties = {list.getString("value"), list.getInt("tag_id")+""};
                tags.add(properties);
            }
            list.close();
        } catch (SQLException ex) {
            Logger.getLogger(ArmadilloDB.class.getName()).log(Level.SEVERE, null, ex);
        }
        return tags;
    }

    /** Returns a tag_id of the required tag, or -1 if tag is not found in the DB. *
     * @param value
     * @param tagtypeID
     * @return
     */
    public int getTag(String value, int tagtypeID)
    {
        int tagID = -1;
        try {
            ResultSet list = this.executeQry("SELECT * FROM 'tag' WHERE tagtype_id = " + tagtypeID + ";");
            while (list.next())
            {
                if (list.getString("value").equals(value))
                {
                    tagID = list.getInt("tag_id");
                    return tagID;
                }
            }
            list.close();
        } catch (SQLException ex) {
            Logger.getLogger(ArmadilloDB.class.getName()).log(Level.SEVERE, null, ex);
        }
        return tagID;
    }

    /**
     * Gets the entire music lbrary current stored in the DB
     * @return
     */
    public ArrayList<String[]> getLibrary() {
        String command = "SELECT * FROM 'track' ORDER BY 'track_id' ;";
        ArrayList<String[]> list = getSelectedSongs(command);
        return list;
    }
    
    /**
     * Gets the contents of a given playlist
     * @param playlistID
     * @return
     */
    public ArrayList<String[]> getPlaylistContents(int playlistID) {
        String command = "SELECT * FROM 'playlist_track' NATURAL JOIN 'track' WHERE playlist_id = "+playlistID+";";
        ArrayList<String[]> playlistContents = getSelectedSongs(command);
        return playlistContents;
    }

    /**
     * Gets the track associated with a track id
     * @param track_id
     * @return
     */
    public String[] getTrack(int track_id) {
        String command = "SELECT * FROM 'track' WHERE track_id = " + track_id + ";";
        ArrayList<String[]> track = getSelectedSongs(command);
        if (track.size() == 0) {
            return null;
        } else {
            return track.get(0);
        }
    }



    /**
     * Gets the tracks associated with a given tag
     * @param tag_id
     * @return
     */
    public ArrayList<String[]> getTracksByTag(int tag_id)
    {
        String command = "SELECT * FROM 'track_tags' NATURAL JOIN 'track' WHERE tag_id = " + tag_id + ";";
        ArrayList<String[]> tracks = getSelectedSongs(command);
        return tracks;
    }

    /** Returns contents of the Library as ArrayList< String [ 7 ] > where<br />
        [0] - TITLE<br />
        [1] - ARTIST<br />
        [2] - ALBUM<br />
        [3] - GENRE<br />
        [4] - LENGTH (DURATION)<br />
        [5] - LOCATION (PATH)<br />
        [6] - TRACK_ID.
     *
     * @param command
     * @return
     */
    public ArrayList<String[]> getSelectedSongs (String command)
    {
        ArrayList<String[]> list = new ArrayList<String[]>();
        try {
            // GET ALL TRACKS
            ResultSet list1 = this.executeQry(command);

            // STORE ID, LENGTH, LOCATION OF EACH TRACK IN list
            while (list1.next()) {
                // FIELDS ARE: TITLE, ARTIST, ALBUM, GENRE, LENGTH, LOCATION, ID
                String[] track = new String[7];
                track[6] = "" + list1.getInt("track_id");
                track[5] = list1.getString("location");
                track[4] = list1.getString("duration");
                list.add(track);
            }
            list1.close();



            // GET ALL TAGS ASSIGNED TO EACH TRACK
            for (int i=0; i<list.size(); i++)
            {
                ResultSet list2 = this.executeQry("SELECT * FROM 'track_tags' WHERE track_id = "+list.get(i)[6]+";");
                ArrayList<Integer> tagIDs = new ArrayList<Integer>();
                while (list2.next())
                {
                  int tag = list2.getInt("tag_id");
                  if (tag != -1)
                    tagIDs.add(tag);
                  // TODO: what if there are no tags for given track - returns -1
                }
                list2.close();

                // GET TITLE, ARTIST, ALBUM, GENRE OF EACH TRACK AND STORE IN list
                //System.out.println("SIZE: "+tagIDs.size());
                for (int j=0; j<tagIDs.size(); j++)
                {
                    ResultSet listA = this.executeQry("SELECT * FROM 'tag' WHERE tag_id = "+tagIDs.get(j)+";");
                    int id = listA.getInt("tagtype_id");
                    
                    if (id == 1) // TITLE
                        list.get(i)[0] = listA.getString("value");
                    if (id == 2) // ARTIST
                        list.get(i)[1] = listA.getString("value");
                    if (id == 3) // ALBUM
                        list.get(i)[2] = listA.getString("value");
                    if (id == 4) // GENRE
                    {
                        //System.out.println("Found genre "+listA.getString("value")+" For track "+list.get(i)[6]);
                        list.get(i)[3] = listA.getString("value");
                    }
                    listA.close();
                }
            }
        }
        catch (SQLException ex) {
            Logger.getLogger(ArmadilloDB.class.getName()).log(Level.SEVERE, null, ex);
        }
        return list;
    }

    /** Add track (ID) to playlist (ID). *
     * @param playlistID
     * @param trackID
     * @param position
     */
    public void addTrackToPlaylist(int playlistID, int trackID, int position)
    {
        try
        {
            PreparedStatement prepStmt = conn.prepareStatement("INSERT INTO playlist_track VALUES (?, ?, ?);");
            prepStmt.setInt(1, playlistID);
            prepStmt.setInt(2, trackID);
            prepStmt.setInt(3, position);
            prepStmt.addBatch();
            prepStmt.executeBatch();
        }
        catch (SQLException e)
        {
            System.err.println(e);
        }
    }

    /** Remove all songs from playlist (ID) *
     * @param playlistID
     */
    public void clearPlaylist (int playlistID)
    {
        try {
            this.executeStmt("DELETE FROM 'playlist_track' WHERE playlist_id = " + playlistID + ";");
        } catch (SQLException ex) {
            Logger.getLogger(ArmadilloDB.class.getName()).log(Level.SEVERE, null, ex);
        }
    }

    /** Update playlist (ID) *
     * @param trackIDs
     * @param playlistID
     */
    public void updatePlaylist(ArrayList<Integer> trackIDs, int playlistID)
    {
        clearPlaylist(playlistID);
        for (int i=0; i<trackIDs.size(); i++)
            addTrackToPlaylist(playlistID, trackIDs.get(i), i);
    }

    /** Add new playlist to DB containing all tracks from the given ArrayList< Integer >.*
     * @param name
     * @param tracks
     */
    public void addPlaylist(String name, ArrayList<Integer> tracks)
    {        
        this.insertPlaylist(name, 0);
        int playlistID = getPlayListID(name);
        for (int i=0; i<tracks.size(); i++)
            this.addTrackToPlaylist(playlistID, tracks.get(i), i);
    }

    /** Add new mood model to DB. *
     * @param tagID
     * @param description
     * @param frequency
     * @param bpm
     */
    public void addMoodModel(int tagID, String description, int bpm, int frequency)
    {
        try
        {
            PreparedStatement prepStmt = conn.prepareStatement("INSERT INTO moodmodel VALUES (?, ?, ?, ?);");
            prepStmt.setInt(1, tagID);
            prepStmt.setString(2, description);
            prepStmt.setInt(3, bpm);
            prepStmt.setInt(4, frequency);
            prepStmt.addBatch();
            prepStmt.executeBatch();
        }
        catch (SQLException e)
        {
            System.err.println(e);
        }
    }

    /** Remove given playlist from DB. *
     * @param id
     */
    public void removePlaylist(int id)
    {
        try {
            this.executeStmt("DELETE FROM 'playlist_track' WHERE playlist_id = " + id + ";");
            this.executeStmt("DELETE FROM 'playlist' WHERE playlist_id = " + id + ";");
        } catch (SQLException ex) {
            Logger.getLogger(ArmadilloDB.class.getName()).log(Level.SEVERE, null, ex);
        }
    }

    /** Remove track (ID) from playlist (ID). *
     * @param trackID
     * @param position
     * @param playlistID
     */
    public synchronized void removeTrackFromPlaylist(int trackID, int playlistID, int position)
    {
        try
        {
            //System.out.println("Removing track from playlist at pos "+position);
            this.executeStmt("DELETE FROM 'playlist_track' WHERE (playlist_id = " + playlistID + ") AND (track_id = "+trackID+") AND (position = "+position+");");
            //showAllPlaylistTracks();
            updateTrackPosition(playlistID, position);
            //showAllPlaylistTracks();
        }
        catch (SQLException e)
        {
            System.err.println(e);
        }
    }

    /** Change (DECREMENT BY 1) track's (ID) position in a playlist (ID). *
     * @param playlistID
     * @param position
     */
    public synchronized void updateTrackPosition(int playlistID, int position)
    {
        try {
            this.executeStmt("UPDATE 'playlist_track' SET position = position-1 WHERE (playlist_id = " + playlistID + ") AND (position > " + position + ");");

        } catch (SQLException ex) {
            Logger.getLogger(ArmadilloDB.class.getName()).log(Level.SEVERE, null, ex);
        }

    }

    /** Remove track (ID) from the Library (and all playlists). *
     * @param id
     */
    public synchronized void removeTrack(int id)
    {
        try {
            // remove track from any playslists
            ResultSet list = this.executeQry("SELECT * FROM 'playlist_track' WHERE track_id = "+id+";");
            while (list.next())
            {
                removeTrackFromPlaylist(id, list.getInt("playlist_id"), list.getInt("position"));
            }
            // remove all tags unless used by other tracks
            this.detag(id);
            this.executeStmt("DELETE FROM 'track' WHERE track_id = " + id + ";");
        } catch (SQLException ex) {
            Logger.getLogger(ArmadilloDB.class.getName()).log(Level.SEVERE, null, ex);
        }
    }

    /** Edit track tag. Returns 0 if successful, -1 if not. *
     * @param trackID
     * @param tagtypeID
     * @param newTag
     */
    public synchronized void editTrackTag(int trackID, int tagtypeID, String newTag)
    {
        try {
        // get the old tag's ID
        int oldTagID = -1;
        ResultSet list = this.executeQry("SELECT * FROM 'track_tags' NATURAL JOIN 'tag' WHERE track_id = " + trackID +" AND tagtype_id = "+tagtypeID+" ;");
        while (list.next())
                oldTagID = list.getInt("tag_id");
        // de-tag the track
        this.executeStmt("DELETE FROM 'track_tags' WHERE track_id = " + trackID +" AND tag_id = "+oldTagID+" ;");
        // first check if the new tag already exists in db, if not create it
        int newTagID = getTag(newTag, tagtypeID);
        if (newTagID == -1)
            newTagID = insertTag(newTag, tagtypeID);
        // then tag the track
        tagTrack(newTagID, trackID);
        // check if the old tag is used anywhere else, if not remove it from db
        ResultSet list2 = this.executeQry("SELECT * FROM 'track_tags' WHERE tag_id = " + oldTagID +" ;");
        if(list2.isClosed())
            this.executeStmt("DELETE FROM 'tag' WHERE tag_id = " + oldTagID +" ;");
        }
        catch (SQLException ex) {
            Logger.getLogger(ArmadilloDB.class.getName()).log(Level.SEVERE, null, ex);
        }
    }

    /** Edit tag. *
     * @param tagID
     * @param newValue
     */
    public void editTag(int tagID, int tagtypeID, String newValue)
    {
        try {
            newValue = newValue.replace("'", "''");
            // check if there is a tag with the same value already and merge both tags
            ResultSet duplicates = this.executeQry("SELECT * FROM 'tag' WHERE value = '" + newValue +"' AND tagtype_id = "+tagtypeID+" ;");
            if (duplicates.isClosed())
            {
                this.executeStmt("UPDATE 'tag' SET value = '" + newValue +"' WHERE tag_id = "+tagID+" ;");
                //System.out.println("updatng tag, no duplicates");
            }
            /*else
            {
                while(duplicates.next())
                {
                    System.out.println("Same tag exisis");
                    int jointID = duplicates.getInt("tag_id");
                    this.executeStmt("DELETE FROM 'navTree' WHERE tag_id = "+tagID+" ;");
                    ResultSet dup = this.executeQry("SELECT * FROM 'track_tags' WHERE tag_id = "+tagID+" ;");
                    if (dup.isClosed())
                        this.executeStmt("UPDATE 'track_tags' SET tag_id = "+jointID+" WHERE tag_id = "+tagID+" ;");
                }
             }*/
        }
        catch (SQLException ex) {
            Logger.getLogger(ArmadilloDB.class.getName()).log(Level.SEVERE, null, ex);
        }
    }

    /** Returns the bpm (int) of the given track ID *
     * @param trackID
     * @return
     */
    public int getBPM(int trackID)
    {
        int bpm = -1;
        try {
            ResultSet list = this.executeQry("SELECT * FROM 'track_tags' NATURAL JOIN 'tag' WHERE track_id = " + trackID + " AND tagtype_id = 5 ;");
            while (list.next())
                bpm = java.lang.Math.round(Float.parseFloat(list.getString("value")));
        } catch (SQLException ex) {
            Logger.getLogger(ArmadilloDB.class.getName()).log(Level.SEVERE, null, ex);
        }
        return bpm;
    }

    /** Return a list of track IDs of the bpm wanted (-10/+9). Returns -1 if no tracks match the given bpm. *
     * @param bpm
     * @return
     */
    public ArrayList<Integer> getTracks(int bpm)
    {
        ArrayList<Integer> trackIDs = new  ArrayList<Integer>();
        // set the range of bpm's we are looking for || CHANGED FROM +-10 to -10/+9, so no overlap
        int high = bpm+10;
        int low = bpm-10;
        try {
            for (int j =low; j<high; j++){
                ResultSet list = this.executeQry("SELECT track_id FROM 'track_tags' NATURAL JOIN 'tag' WHERE tagtype_id = 5 AND value = "+j+" ;");
                while (list.next()) {
                        trackIDs.add(list.getInt("track_id"));
                }
            }
            if (trackIDs.size() == 0)
                {
                    trackIDs.add(-1);
                    return trackIDs;
                }
        } catch (SQLException ex) {
            Logger.getLogger(ArmadilloDB.class.getName()).log(Level.SEVERE, null, ex);
        }
        return trackIDs;
    }

    /** Return a list of track IDs of the bpm wanted (-10/+9). Returns -1 if no tracks match the given bpm. *
     * @param bpm
     * @return
     */
    public ArrayList<Integer> getTracksWithMatchingBPM(int bpm)
    {
        ArrayList<Integer> trackIDs = new  ArrayList<Integer>();

        try {
                ResultSet list = this.executeQry("SELECT track_id FROM 'track_tags' NATURAL JOIN 'tag' WHERE tagtype_id = 5 AND value = "+bpm+" ;");
                while (list.next()) {
                        trackIDs.add(list.getInt("track_id"));
                }
           
            if (trackIDs.size() == 0)
                {
                    trackIDs.add(-1);
                    return trackIDs;
                    }
        } catch (SQLException ex) {
            Logger.getLogger(ArmadilloDB.class.getName()).log(Level.SEVERE, null, ex);
        }
        return trackIDs;
    }

    /**
     *  Return true if this music file has already been imported, else false.
     * @param path
     * @return
     */
    public boolean exists(String path)
    {        
        try {
            path = path.replace("'", "''");
            ResultSet list = this.executeQry("SELECT * FROM 'track' WHERE location = '"+path+"';");
            while (list.next()) {
               return true;
            }
        } catch (SQLException ex) {
            Logger.getLogger(ArmadilloDB.class.getName()).log(Level.SEVERE, null, ex);
        }
        return false;
    }


    /**
     *  Return true if this music file has already been imported, else false.
     * @param trackID
     * @return
     */
    public boolean exists(int trackID)
    {
        try {
            ResultSet list = this.executeQry("SELECT * FROM 'track' WHERE track_id = "+trackID+";");
            while (list.next()) {
               return true;
            }
        } catch (SQLException ex) {
            Logger.getLogger(ArmadilloDB.class.getName()).log(Level.SEVERE, null, ex);
        }
        return false;
    }

    /**
     * Searchs the database for tracks linked with the given keywords
     * @param keywords
     * @return
     */
    public int search(String[] keywords)
    {
        ArrayList<Integer> tracksfound = new ArrayList<Integer>();
        ArrayList<Integer> result = new ArrayList<Integer>();
        this.clearPlaylist(2);
        try {
            for (int k=0; k<keywords.length; k++)
            {
                keywords[k] = keywords[k].replace("'", "''");
                ResultSet taglist = this.executeQry("SELECT * FROM 'tag' NATURAL JOIN 'track_tags' WHERE value LIKE '%" + keywords[k] + "%' ;");
                if (taglist.isClosed())
                {
                    new DialogUI(7).setVisible(true);
                    return 2;
                }
                while (taglist.next()) {
                    if (k == 0)
                    {
                        if (!(tracksfound.contains(taglist.getInt("track_id"))))
                            tracksfound.add(taglist.getInt("track_id"));
                    }
                    else
                    {
                        if ((tracksfound.contains(taglist.getInt("track_id"))) && !(result.contains(taglist.getInt("track_id"))))
                            result.add(taglist.getInt("track_id"));
                    }
                }                
            }
            if (keywords.length == 1)
                result = tracksfound;
            for (int i = 0; i < result.size(); i++)
                this.addTrackToPlaylist(2, result.get(i), i+1);
            
        } catch (SQLException ex) {
            Logger.getLogger(ArmadilloDB.class.getName()).log(Level.SEVERE, null, ex);
        }
        return 2; // this is just to make the table wait for the results before reloading
    }

    // TESTING ONLY
    /**
     * shows all the tags currently in the library
     */
    public void showAllTags()
    {
        try {
            ResultSet list = this.executeQry("SELECT * FROM 'tag';");
            while (list.next()) {
                System.out.println("TAG_ID: " + list.getInt("tag_id"));
                System.out.println("TAG_VALUE: " + list.getString("value"));
                System.out.println("TAGTYPE_ID: " + list.getInt("tagtype_id"));
            }
        } catch (SQLException ex) {
            Logger.getLogger(ArmadilloDB.class.getName()).log(Level.SEVERE, null, ex);
        }
    }

    // TESTING ONLY
    /**
     * shows all the playlist tracks currently in the library
     */
    public void showAllPlaylistTracks()
    {
        System.out.println("SHOWING ALL PLAYLIST TRACKS...");
        try {
            ResultSet list = this.executeQry("SELECT * FROM 'playlist_track';");
            while (list.next()) {
                System.out.println("playlistID: " + list.getInt("playlist_id"));
                System.out.println("trackID: " + list.getInt("track_id"));
                System.out.println("position: " + list.getInt("position"));
            }
        } catch (SQLException ex) {
            Logger.getLogger(ArmadilloDB.class.getName()).log(Level.SEVERE, null, ex);
        }
    }

    /**
     * returns all songs which still need to be analyized
     * @return
     */
    public int[] getTracksToAnalyse()
    {
        ArrayList<Integer> result = new ArrayList<Integer>();
        try {
            ResultSet list = this.executeQry("SELECT * FROM 'track' WHERE analysed = 'false';");
            while (list.next())
                result.add(list.getInt("track_id"));
        } catch (SQLException ex) {
            Logger.getLogger(ArmadilloDB.class.getName()).log(Level.SEVERE, null, ex);
        }
        int[] list = new int[result.size()];
        for (int i=0; i<list.length; i++)
            list[i] = result.get(i);
        return list;

    }

    /**
     * checks if there are any tracks left to analyse
     * @return
     */
    public boolean anyTracksToAnalyse()
    {
        try {
            ResultSet list = this.executeQry("SELECT * FROM 'track' WHERE analysed = 'false';");
            while (list.next())
                return true;
        } catch (SQLException ex) {
            Logger.getLogger(ArmadilloDB.class.getName()).log(Level.SEVERE, null, ex);
        }
        return false;
    }

    /**
     * Store the doubles generated from the analysis of a song in the DB
     * @param track_id
     * @param array
     */
    public void storeFPDoubles(int track_id, double[] array)
    {
        try {
            String command = "?";
            for (int i=0; i<50; i++)
                command = command.concat(", ?");
            PreparedStatement prepStmt = conn.prepareStatement("INSERT INTO fingerprintDoubles VALUES ("+command+");");
            prepStmt.setInt(1, track_id);
            for (int i=0; i<50; i++)
                prepStmt.setDouble(i+2, array[i]);
            prepStmt.addBatch();
            prepStmt.executeBatch();
        } catch (SQLException ex) {
            Logger.getLogger(ArmadilloDB.class.getName()).log(Level.SEVERE, null, ex);
        }
    }

    /**
     * Stores the integers generated from the music analysis in the DB
     * @param track_id
     * @param array
     */
    public void storeFPInts(int track_id, int[] array)
    {
        try {
            String command = "?";
            for (int i=0; i<50; i++)
                command = command.concat(", ?");
            PreparedStatement prepStmt = conn.prepareStatement("INSERT INTO fingerprintInts VALUES ("+command+");");
            prepStmt.setInt(1, track_id);
            for (int i=0; i<50; i++)
                prepStmt.setInt(i+2, array[i]);
            prepStmt.addBatch();
            prepStmt.executeBatch();
        } catch (SQLException ex) {
            Logger.getLogger(ArmadilloDB.class.getName()).log(Level.SEVERE, null, ex);
        }
    }

    /**
     * gets the music analysis doubles associated with a track
     * @param track_id
     * @return
     */
    public double[] getFPDoubles(int track_id)
    {
        double[] values = new double[50];
        try {
            ResultSet list = this.executeQry("SELECT * FROM 'fingerprintDoubles' WHERE track_id = "+track_id+" ;");
            while (list.next()) {
                for (int j=1; j<51; j++)
                    values[j-1] = list.getDouble(j+1);
            }
        }
        catch (SQLException ex) {
            Logger.getLogger(ArmadilloDB.class.getName()).log(Level.SEVERE, null, ex);
        }
        return values;
    }

    /**
     * gets the music analysis integers associated with a track
     * @param track_id
     * @return
     */
    public int[] getFPInts(int track_id)
    {
        int[] values = new int[50];
        try {
            ResultSet list = this.executeQry("SELECT * FROM 'fingerprintInts' WHERE track_id = "+track_id+" ;");
            while (list.next()) {
                for (int j=1; j<51; j++)
                    values[j-1] = list.getInt(j+1);
            }
        }
        catch (SQLException ex) {
            Logger.getLogger(ArmadilloDB.class.getName()).log(Level.SEVERE, null, ex);
        }
        return values;
    }

    // TESTING ONLY
    /**
     * shows the fingerprint data of a given track
     * @param track_id
     */
    public void showFingerprints(int track_id)
    {
        System.out.println("Showing fingerprints doubles for track_id = "+track_id);
        double[] values = this.getFPDoubles(track_id);
        for (int i=0; i<50; i++)
            System.out.print("  "+i+": "+values[i]);
        System.out.println("");
        System.out.println("Showing fingerprints ints for track_id = "+track_id);
        int[] values2 = this.getFPInts(track_id);
        for (int i=0; i<50; i++)
            System.out.print("  "+i+": "+values2[i]);
        System.out.println("");
        
    }

    /**
     * removes all tags from a track
     * @param track_id
     **/
    private void detag(int id) {
        ArrayList<Integer> tags = new ArrayList<Integer>();
        try {
            ResultSet list3 = this.executeQry("SELECT * FROM 'track_tags' WHERE track_id = " + id + " ;");
            while (list3.next())
                tags.add(list3.getInt("tag_id"));
            list3.close();
            for (int y=0; y<tags.size(); y++)
            {
                this.executeStmt("DELETE FROM 'track_tags' WHERE track_id = " + id + " AND tag_id = " + tags.get(y) + " ;");
                ResultSet list2 = this.executeQry("SELECT * FROM 'track_tags' WHERE tag_id = " + tags.get(y) + " ;");
                if (list2.isClosed() == true) 
                    this.executeStmt("DELETE FROM 'tag' WHERE tag_id = " + tags.get(y) + " ;");
            }
        } catch (SQLException ex) {
            Logger.getLogger(ArmadilloDB.class.getName()).log(Level.SEVERE, null, ex);
        }
    }

    /**
     * Saves the current queue of tracks to analysis in the db
     * @param queue
     */
    public void saveQueue(LinkedList<int[]> queue)
    {
        for (int i=0; i<queue.size(); i++)
        {
            int[] batch = queue.get(i);
            for (int j=0; j<batch.length; j++)
            {
                try {
                    PreparedStatement prepStmt = this.conn.prepareStatement("INSERT INTO importQueue VALUES (?, ?);");
                    prepStmt.setInt(1, (i + 1));
                    prepStmt.setInt(2, batch[j]);
                    prepStmt.addBatch();
                    prepStmt.executeBatch();
                } catch (SQLException ex) {
                    Logger.getLogger(ArmadilloDB.class.getName()).log(Level.SEVERE, null, ex);
                }
            }
        }
    }

    /**
     * Gets the current queue of tracks to analysis from the db
     * @return
     */
    public LinkedList<int[]> getQueue()
    {
        LinkedList<int[]> queue = new LinkedList<int[]>();
        int noOfBatches = 0;
        try {
            ResultSet noBatches = this.executeQry("SELECT batch_no FROM 'importQueue' ;");
            while (noBatches.next()) {
                if (noBatches.getInt("batch_no") > noOfBatches)
                    noOfBatches = noBatches.getInt("batch_no");
            }
            for (int i=1; i<=noOfBatches; i++)
            {
                int batchSize =0, counter = 0;
                ResultSet list = this.executeQry("SELECT * FROM 'importQueue' WHERE batch_no = "+i+" ;");
                while (list.next())
                    batchSize++;
                int[] batch = new int[batchSize];
                ResultSet list2 = this.executeQry("SELECT * FROM 'importQueue' WHERE batch_no = "+i+" ;");
                while (list2.next())
                {
                    batch[counter] = list2.getInt("track_id");
                    counter++;
                }
                queue.addLast(batch);
            }
        } catch (SQLException ex) {
            Logger.getLogger(ArmadilloDB.class.getName()).log(Level.SEVERE, null, ex);
        }
        return queue;
    }

    // TESTING ONLY
    /**
     * shows the current queue of tracks to analysis
     **/
    public void showQueue()
    {
        System.out.println("Showing queue");
        try {
            ResultSet list = this.executeQry("SELECT * FROM 'importQueue' ;");
            while (list.next()) {
                System.out.println("batch_no "+list.getInt("batch_no")+" track_id "+list.getInt("track_id"));
            }
        }
        catch (SQLException ex) {
            Logger.getLogger(ArmadilloDB.class.getName()).log(Level.SEVERE, null, ex);
        }
    }

    /**
     * Stores the average decibels associated with a track
     * @param trackID
     * @param averageDec
     */
    public void storeAverageDec(int trackID, double averageDec) {
        try
        {
            if (averageDec <= 0)
            {
                averageDec = 0.0;
            }
            PreparedStatement prepStmt = conn.prepareStatement("INSERT INTO fingerprint VALUES (?, ?);");
            prepStmt.setInt(1, trackID);
            prepStmt.setDouble(2, averageDec);
            prepStmt.addBatch();
            prepStmt.executeBatch();
        }
        catch (SQLException e)
        {
            System.err.println(e);
        }
    }

    /**
     * gets the average decibels associated with a track
     * @param trackID
     * @return
     */
    public double getAverageDec(int trackID)
    {
        try {
            ResultSet list = this.executeQry("SELECT averageDec FROM 'fingerprint' WHERE track_id = "+trackID+" ;");
            while (list.next()) {
                return list.getDouble("averageDec");
            }
        }
        catch (SQLException ex) {
            Logger.getLogger(ArmadilloDB.class.getName()).log(Level.SEVERE, null, ex);
        }
        return -1;
    }

    /**
     * clears the current song analysis queue
     */
    public void clearQueueTable() {
        try {
            this.executeStmt("drop table if exists 'importQueue' ;");
            this.executeStmt("CREATE TABLE importQueue (batch_no INTEGER NOT NULL, track_id INTEGER NOT NULL);");
        }
        catch (SQLException ex) {
            Logger.getLogger(ArmadilloDB.class.getName()).log(Level.SEVERE, null, ex);
        }
    }
}
