/*   This file is part of Armadillo Music Player.
*
*   Armadillo is free software: you can redistribute it and/or modify
*   it under the terms of the GNU General Public License as published by
*   the Free Software Foundation, either version 3 of the License, or
*   (at your option) any later version.
*
*   Armadillo is distributed in the hope that it will be useful,
*   but WITHOUT ANY WARRANTY; without even the implied warranty of
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*   GNU General Public License for more details.
*
*   You should have received a copy of the GNU General Public License
*   along with Armadillo.  If not, see <http://www.gnu.org/licenses/>.
*/

package armadilloUI;

import java.io.*;
import java.nio.*;
import org.gstreamer.*;
import org.gstreamer.Buffer;
import org.gstreamer.Structure;
import org.gstreamer.elements.AppSink;
import org.gstreamer.elements.PlayBin;


/**
 * This class is the manager for all track analysis.  This class will work out
 * the BPM or beat of a track, and also controls when the track should have
 * additional processing performed to execute the frequency analysis required
 * to fingerprint the track.
 */
public class BPM
{
    private int pMusic = 0;
    private int freq_N; 
    private int nearest_fft_value;
    private float[] instantE;
    private int ListCount = -1;
    private float[] freq2;
    private int rate;
    private int num_samples;
    private int channels = 0;
    private int count = 0;
    private BPM_BIN BPMbin[] = new BPM_BIN[100];
    private Bin moods = new Bin();
    private int beat;
    private double[] startA = new double[5];
    private double[] endA   = new double[5];
    private int BPMcount = 0;
    private int BPMcountprev = 1;
    private int chain = 0;
    private double onesec_bufcount;
    private int highBPMinterval = 1;
    private int lowBPMinterval = 0;
    /**
    * Create only useful BPM bins (will give BPM between ~60 to ~300
    * on first call of BPM class
    **/
    public BPM() {
        ArmadilloLog.LogOperation("Constructing BPM object");
    }

    /**
     * Initialises all the BPM bins
     */
    public void initialiseBPMbins()
    {
        for (int i = lowBPMinterval; i <= highBPMinterval; i++)
        {
            BPMbin[i] = new BPM_BIN();
        }
    }

    /**
     * Deals with individual buffers, and saves the data into the correct
     * array, and passes information to be fingerprinted.
     **/ 
    private boolean process_buffer(Buffer buf, FingerPrint this_song, boolean finger)
    {
        // Figure out the format of this buffer
        Caps bufcaps = buf.getCaps();
        int size = buf.getSize();
        if (!bufcaps.isFixed())
        {
            System.err.println("Error: Buffer should have fixed Caps.");
            return false;
        }
        Structure bcaps = bufcaps.getStructure(0);
        rate = bcaps.getInteger("rate");
        channels = bcaps.getInteger("channels");
        if (channels > 2)
        {
            System.err.println("Error: Too many channels " + channels);
            return false;
        }
        // Get the raw data out of the buffer
        ByteBuffer bbuf = buf.getByteBuffer();
        FloatBuffer fbuf = bbuf.asFloatBuffer();
        float[] samples = new float[size / 4];
        fbuf.get(samples);
        num_samples = samples.length / channels;
        freq_N = (num_samples / 3);
        nearest_fft_value = 32;
        while (nearest_fft_value < freq_N)
        {
            nearest_fft_value = (nearest_fft_value*2);
        }
        freq2 = new float[nearest_fft_value];
        ListCount++;
        if (ListCount == highBPMinterval) ListCount = 0;
        for (int i = 0; i < num_samples; i++)
        {  
            pMusic++;
            float temp_sample=0;
            float actual_sample=0;
            switch (channels)
            {
                case 1:
                {
                    temp_sample = (samples[i]*samples[i]);
                    actual_sample = (samples[i]);
                    break;
                }
                case 2:
                {
                    temp_sample = (samples[2*i]*samples[2*i]) + (samples[(2*i)+1]*samples[(2*i)+1]);
                    actual_sample = (samples[2*i] + samples[(2*i)+1]) / 2;
                    break;
                }
                default:
                {
                    ArmadilloLog.LogError("BPM object - more than 2 channels");
                    System.err.println("Greater than 2 channels!");
                    System.exit(-1);
                }
            }
            if (finger == true)
            {
                freq2[count] = actual_sample;
                if (count < (nearest_fft_value-1))
                {
                    count++;
                } 
                else 
                {
                    ArmadilloLog.LogError("BPM object - invalid song type");
                    System.err.println("Invalid song type: mood model may be incorrect"); 
                }
                if ((i+1) % freq_N == 0)
                {
                    this_song.run(freq2, rate, freq_N);
                    count = 0;
                    freq2 = new float[nearest_fft_value];
                }
            }
            if (i > 0) temp_sample += instantE[ListCount];
            instantE[ListCount] = temp_sample;
        }
        return true;
    }

    /**
     * Sets up variables required for the calucations due to follow.
     * @param args
     */
    public void TestRun(String args)
    {
        AppSink asTest = (AppSink)ElementFactory.make("appsink", "testsink");
        Caps caps = new Caps("audio/x-raw-float, width=32");
        asTest.setCaps(caps);
        PlayBin playbinTest = new PlayBin("AudioPlayer");
        playbinTest.setInputFile(new File(args));
        playbinTest.setAudioSink(asTest);
        playbinTest.setVideoSink(ElementFactory.make("fakesink", "videosink"));
        playbinTest.useClock(null);
        playbinTest.setState(State.PLAYING);
        playbinTest.getState();
        Buffer bufTest;
        int TestCounter = 0;
        int Prev_num_samples = 0;
        while (TestCounter < 10)
        {
            if ((bufTest = asTest.pullBuffer()) != null)
            {
                Caps bufcaps = bufTest.getCaps();
                int size = bufTest.getSize();
                Structure bcaps = bufcaps.getStructure(0);
                rate = bcaps.getInteger("rate");
                channels = bcaps.getInteger("channels");
                ByteBuffer bbuf = bufTest.getByteBuffer();
                FloatBuffer fbuf = bbuf.asFloatBuffer();
                float[] samples = new float[size / 4];
                fbuf.get(samples);
                num_samples = samples.length / channels;
                bufTest.dispose();

                if (num_samples > Prev_num_samples)
                {
                    freq_N = (num_samples / 3);
                    nearest_fft_value = 32;
                    while (nearest_fft_value < freq_N)
                    {
                        nearest_fft_value = (nearest_fft_value*2);
                    }
                    onesec_bufcount = (double)rate/(double)num_samples;
                    double testInterval = 301;
                    lowBPMinterval = 0;
                    while (testInterval > 280)
                    {
                        lowBPMinterval++ ;
                        testInterval = (onesec_bufcount / (double) lowBPMinterval) * (double)60;
                    }
                    highBPMinterval = (int)onesec_bufcount;
                }
                Prev_num_samples = num_samples;
            }
            TestCounter++;
        }
        playbinTest.setState(State.NULL);
        playbinTest.getState();
        playbinTest.dispose();
    }

    void setAn(double timer_buffers, double onesec_bufcount, int factor)
    {
        double temp = timer_buffers/6;
        for (int i = 0 ; i < 5; i++)
        {
            startA[i] = temp - factor*(onesec_bufcount);
            endA[i]   = temp + factor*(onesec_bufcount);
            temp = temp * 2;
        }
    }

    /**
     * Sets the moments of data to be fingerprinted.
     * @param timer_buffers
     * @param onesec_bufcount
     * @param speed
     **/
    public void setSpeeds(double timer_buffers,double onesec_bufcount, int speed)
    {
        switch (speed)
        {
            case 1:
            {
                if (timer_buffers >= ((double)30 * onesec_bufcount))
                {
                    setAn(timer_buffers, onesec_bufcount, 2);
                }
                break;
            }
            case 2:
            {
                if (timer_buffers >= ((double)50 * onesec_bufcount))
                {
                    setAn(timer_buffers, onesec_bufcount, 4);
                }
                break;
            }
            case 3:
            {
                if (timer_buffers >= ((double)120 * onesec_bufcount))
                {
                    setAn(timer_buffers, onesec_bufcount, 8);
                }
                break;
            }
        }        
    }

    /**
     * Sets up gstreamer, begins playback, and feeds buffers to process_buffers.
     * It also controls the evaluation of the BPM
     * @param args
     * @param speed
     * @param override
     * @return Whether to save data into the database (boolean)
     */
    @SuppressWarnings("empty-statement")
    public boolean run(String[] args, int speed, boolean override)
    {
        args = Gst.init("BPM", args);
        FingerPrint this_song = new FingerPrint();
        TestRun(args[0]);
        initialiseBPMbins();
        instantE = new float[highBPMinterval];
        AppSink as = (AppSink)ElementFactory.make("appsink", "testsink");
        Caps caps = new Caps("audio/x-raw-float, width=32");
        as.setCaps(caps);
        PlayBin playbin = new PlayBin("AudioPlayer");
        playbin.setInputFile(new File(args[0]));
        playbin.setAudioSink(as);
        playbin.setVideoSink(ElementFactory.make("fakesink", "videosink"));
        playbin.useClock(null);
        playbin.setState(State.PLAYING);
        playbin.getState();
        Buffer buf;
        int buf_count = 0;
        long timer = playbin.queryDuration(Format.TIME);
        timer = timer / 1000000000;
        onesec_bufcount = (double)rate/(double)num_samples;
        double timer_buffers = onesec_bufcount * (double)timer;
        double V = 0;
        if (timer_buffers < ((double)30  * onesec_bufcount))
        {
            override = true;
            new DialogUI(20).setVisible(true);
            return false;
        }
        if (timer_buffers < ((double)120 * onesec_bufcount)) speed = 2;
        if (timer_buffers < ((double)50  * onesec_bufcount)) speed = 1;
        setSpeeds(timer_buffers, onesec_bufcount, speed);
        while (((buf = as.pullBuffer()) != null) && (override != true))
        {
            boolean finger = false;
            long t = playbin.queryPosition(Format.TIME);
            t = t/1000000000;
            for (int i = 0 ; i < 5; i++)
            {
                if ((buf_count > startA[i] ) && (buf_count < endA[i])) finger = true;
            }
            if (t >= timer)
            {
                finger = false;
            }
            if (!process_buffer(buf, this_song, finger))
            {
                break;
            }
            double Vprev = V;
            V = 0;
            float secE = 0;
            if (buf_count >= (int)(onesec_bufcount -1))
            {
                for (int i = 0; i < (int)onesec_bufcount; i++) secE += instantE[i];
                secE = secE / (float)((int)onesec_bufcount);
                for (int i = 0; i < (int)onesec_bufcount; i++)
                {
                    V = V + Math.sqrt((instantE[i] - secE) * (instantE[i] - secE));
                }
                V = V / (int)onesec_bufcount;
                double  C = (0.0025714*Vprev) + 1.5642857;
                countBeat(C, secE);
            } 
            buf.dispose();
            buf_count++;
        }
        this_song.normalize(speed);
        playbin.setState(State.NULL);
        BPM_BIN mode = new BPM_BIN();
        BPM_BIN top5modes[] = new BPM_BIN[6];
        top5modes =  top5(top5modes);
        top5modes =  occurances(mode, top5modes);
        int testdoubleBPM = top5modes[1].BinNumber * 2;
        mode = doubleBPM(testdoubleBPM, mode, top5modes);
        if (mode.BinNumber == 0) mode = top5modes[1];
        moods = this_song.return_bin();
        beat = (int)predominantBeat(mode, top5modes);;
        playbin.setState(State.NULL);
        playbin.getState();
        playbin.dispose();
        return true;
    }

    /**
    *   If the previous energy value is greater than (C * average energy of a second),
    *   it checks this value is a peak amongst its neighbour instant energy values
     *
     * @param C
     * @param secE
     */
    public void countBeat(double C, double secE)
    {
        if (ListCount > 0)
        {
            if (ListCount > 1)
            {
                if ((instantE[ListCount-1] > (C* secE )) && (instantE[ListCount-1] > instantE[ListCount-2]) && (instantE[ListCount-1] > instantE[ListCount]))
                {
                    BPMcount++;
                            if (BPMcount == BPMcountprev) chain++;
                            else chain = 0;
                            if ((BPMcount >= lowBPMinterval) && (BPMcount <=highBPMinterval)) BPMbin[BPMcount].inc(BPMcount, chain, rate, num_samples);
                            BPMcountprev = BPMcount;
                            BPMcount = 0;
                        }
                        else BPMcount++;
                    }
                    else
                    {
                        if ((instantE[0] > (C* secE )) && (instantE[0] > instantE[(highBPMinterval-1)]) && (instantE[0] > instantE[1]))
                        {
                            BPMcount++;
                            if (BPMcount == BPMcountprev) chain++;
                            else chain = 0;
                            if ((BPMcount >= lowBPMinterval) && (BPMcount <=highBPMinterval)) BPMbin[BPMcount].inc(BPMcount, chain, rate, num_samples);
                            BPMcountprev = BPMcount;
                            BPMcount = 0;
                        }
                        else BPMcount++;
                    }
                }
                else
                {
                    if ((instantE[(highBPMinterval-1)] > (C* secE )) && (instantE[(highBPMinterval-1)] > instantE[(highBPMinterval-2)]) && (instantE[(highBPMinterval-1)] > instantE[0]))
                    {
                        BPMcount++;
                        if (BPMcount == BPMcountprev) chain++;
                        else chain = 0;
                        if ((BPMcount >= lowBPMinterval) && (BPMcount <=highBPMinterval)) BPMbin[BPMcount].inc(BPMcount, chain, rate, num_samples);
                        BPMcountprev = BPMcount;
                        BPMcount = 0;
                    }
                    else BPMcount++;
                }
    }
   
    /** 
    * Get the top 5 modes in terms of occurances
     *
     * @param top5modes
     * @return
     */
    public BPM_BIN[] top5(BPM_BIN top5modes[])
    {
        int modebin = 0;
        int modecount = 0;
        for (int j = 1; j <= 5; j++)
        {
            for (int i = lowBPMinterval; i <= highBPMinterval; i++)
            {
                if ((BPMbin[i].counter >= modecount) && (BPMbin[i].discovered == 0))
                {
                    modecount = BPMbin[i].counter;
                    modebin = i;
                    top5modes[j] = BPMbin[i];
                }
            }
            BPMbin[modebin].discovered = 1;
            modecount = 0;
        }
    return top5modes;
    }

    /** 
     * If highest consecutive < 5, then rank just by occurances 
     *
     * @param mode
     * @param top5modes
     * @return
     */
    public BPM_BIN[] occurances(BPM_BIN mode, BPM_BIN top5modes[])
    {
        if ((top5modes[1].consecutive > 4) || (top5modes[2].consecutive > 4) || (top5modes[3].consecutive > 4) || (top5modes[4].consecutive > 4) || (top5modes[5].consecutive > 4))
        {
            for (int k = 1; k <= 5; k++)
            {
                for (int l = 5; l > k; l--)
                {
                    if ((top5modes[l].consecutive > top5modes[l-1].consecutive) || ((top5modes[l].consecutive == top5modes[l-1].consecutive) && (top5modes[l].counter > top5modes[l-1].counter)))
                    {
                        top5modes[0] = top5modes[l];
                        top5modes[l] = top5modes[l-1];
                        top5modes[l-1] = top5modes[0];
                    }
                }
            }
        }
        return top5modes;
    }
    
    /** 
     * Check to find the most predominant beat out of the possible modes
     *
     * @param mode
     * @param top5modes
     * @return
     */
    public double predominantBeat(BPM_BIN mode, BPM_BIN top5modes[])
    {
        double accuratemode = (double)mode.counter * mode.BPMequiv;
        int totaloccurances = mode.counter;
        for (int n = 1; n <= 5; n++)
        {
            if ((top5modes[n].BinNumber == mode.BinNumber - 1) || (top5modes[n].BinNumber == mode.BinNumber + 1))
            {
                accuratemode = accuratemode + ((double)top5modes[n].counter * top5modes[n].BPMequiv);
                totaloccurances = totaloccurances + top5modes[n].counter;
            }
        }

        accuratemode = (accuratemode / (double)totaloccurances) + 0.5;
        return accuratemode;
    }

    /** Check if algorithm has picked up a 'double' BPM *
     * @param testdoubleBPM
     * @param mode
     * @param top5modes
     * @return
     */
    public BPM_BIN doubleBPM(int testdoubleBPM, BPM_BIN mode, BPM_BIN top5modes[])
    {
        if (testdoubleBPM <= 38)
        {
            for (int m = 2; m <= 5; m++)
            {
                if ((top5modes[m].BinNumber == testdoubleBPM) || (top5modes[m].BinNumber == testdoubleBPM - 1) || (top5modes[m].BinNumber == testdoubleBPM + 1))
                {
                    if (top5modes[m].counter > (int)(0.3 * (double)top5modes[1].counter))
                    {
                        mode = top5modes[m];
                        break;
                    }
                }
            }
        }
        return mode;
    }

     /**
     *
     * @return frequency bins
     */
    public Bin get_bin()
    {
        return moods;
    }

    /**
     *
     * @return beats per minute
     */
    public int getBPM()
    {
        return beat;
    }
}