/*   This file is part of Armadillo Music Player.
*
*   Armadillo is free software: you can redistribute it and/or modify
*   it under the terms of the GNU General Public License as published by
*   the Free Software Foundation, either version 3 of the License, or
*   (at your option) any later version.
*
*   Armadillo is distributed in the hope that it will be useful,
*   but WITHOUT ANY WARRANTY; without even the implied warranty of
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*   GNU General Public License for more details.
*
*   You should have received a copy of the GNU General Public License
*   along with Armadillo.  If not, see <http://www.gnu.org/licenses/>.
*/

package armadilloUI;

import javax.swing.*;
import java.awt.*;
import java.awt.geom.Line2D;

/**
 * Draws the graph represntation of a song
 * @author Mark
 */
public class DrawGraph extends JPanel
{
    /**
     *
     */
    protected final int dimSizeX = 500;
    /**
     *
     */
    protected final int dimSizeY = 80;

    /**
     *
     */
    protected final int pad_x = 10;
    /**
     *
     */
    protected final int pad_y = 10;

    /**
     *
     */
    protected Color cLine_out = Color.white;
    /**
     *
     */
    protected Color cLine_in = Color.lightGray;
    /**
     *
     */
    protected Color cSeek = Color.red;

    /**
     *
     */
    protected float innerSize = 0.25f;

    /**
     *
     */
    protected double seek = 0;

    /**
     *
     */
    protected float[] data = null;
    /**
     *
     */
    protected long points = 0;

    /**
     * Constructor, sets up graph border
     */
    public DrawGraph()
    {
        ArmadilloLog.LogOperation("Constructing Draw Graph");
        setBorder(BorderFactory.createLineBorder(Color.black));
    }

    /**
     * returns the preferred size of the graph object
     * @return
     */
    @Override
    public Dimension getPreferredSize()
    {
        return new Dimension(dimSizeX, dimSizeY);
    }

    /**
     * nothing to initialise
     */
    public void init()
    {

    }

    /**
     * updates the graph with new data (from a new song)
     * @param data
     */
    public void update(float[] data)
    {
        this.data = data;
        points = data.length / 2;

        repaint();
    }

    /**
     * moves the position marker to the seek location
     * @param pos
     */
    public void seek(double pos)
    {
        seek = pos;
        repaint();
    }

    /**
     * Reskins the graph with the passed colours
     * @param lineIn
     * @param lineOut
     * @param seekBar
     */
    public void skin(Color lineIn, Color lineOut, Color seekBar)
    {
        cLine_in = lineIn;
        cLine_out = lineOut;
        cSeek = seekBar;

        repaint();
    }

    /**
     * Sets the amount of inner colour the graph has
     * @param size
     */
    public void setInnerSize(float size)
    {
        innerSize = size;

        repaint();
    }

    /**
     * Draws the data on the graph
     * @param g2
     * @param scale
     */
    protected void draw_line(Graphics2D g2, double scale)
    {     
        int width = getWidth() - 2*pad_x;
        int height = getHeight() - 2*pad_y;

        int midpoint = getHeight() / 2;
        int midheight = height / 2;

        for (int x = 0; x < width; ++x)
        {
            int p = new Float((float)x/width * points).intValue();

            double hi_sample = midpoint - data[2*p] * midheight * scale;
            double lo_sample = midpoint - data[2*p+1] * midheight * scale;

            g2.draw(new Line2D.Double(pad_x + x, hi_sample, pad_x + x, lo_sample));
        }
    }

    /**
     * clears the graph
     */
    public void clear()
    {
        data = null;
        seek = 0;
        repaint();
    }
        
    /**
     * set the line type
     * @param g2
     * @param size
     * @param color
     */
    protected void setLine(Graphics2D g2, int size, Paint color)
    {
        g2.setStroke(new BasicStroke (size));
        g2.setPaint(color);
    }
    
    /**
     * paints the component
     * @param g
     */
    @Override
    public void paintComponent(Graphics g)
    {
        super.paintComponent(g);
        Graphics2D g2 = (Graphics2D) g;

        if (data != null)
        {
            // draw data

            setLine(g2, 1, cLine_out);
            draw_line(g2, 1);
            setLine(g2, 1, cLine_in);
            draw_line(g2, innerSize);

            // draw axis
            /*
            setLine(g2, 4, Color.white);
            g2.draw(new Line2D.Double(pad_x, pad_y, pad_x, getHeight() - pad_y));
            */
            // draw seekbar

            double pos = 1 + pad_x + seek * (getWidth() - 2*pad_x);

            setLine(g2, 2, cSeek);
            g2.draw(new Line2D.Double(pos, 0, pos, getHeight()));
        }
    }
}
