/*   This file is part of Armadillo Music Player.
*
*   Armadillo is free software: you can redistribute it and/or modify
*   it under the terms of the GNU General Public License as published by
*   the Free Software Foundation, either version 3 of the License, or
*   (at your option) any later version.
*
*   Armadillo is distributed in the hope that it will be useful,
*   but WITHOUT ANY WARRANTY; without even the implied warranty of
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*   GNU General Public License for more details.
*
*   You should have received a copy of the GNU General Public License
*   along with Armadillo.  If not, see <http://www.gnu.org/licenses/>.
*/

package armadilloUI;

import java.io.File;
import java.nio.ByteBuffer;
import java.nio.FloatBuffer;

import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicInteger;

import org.gstreamer.*;
import org.gstreamer.elements.AppSink;
import org.gstreamer.elements.PlayBin;

class GraphData
{
    // 1 second lookahead is ~40 buffers (38.2...)
    protected int look_ahead = 1;
    protected String file;

    protected float[] data;
    protected float max_data;
    protected float min_data;

    protected int buf_count;
    protected long length;

    protected boolean threaded = false;
    protected int index;
    protected AtomicInteger sync;
    protected boolean term = false;

    public GraphData(String file)
    {
        this.file = file;
    }

    public GraphData(String file, int index, AtomicInteger sync)
    {
        this.file = file;

        threaded = true;
        this.index = index;
        this.sync = sync;
    }

    public float[] getData()
    {
        return data;
    }

    public void setLookAhead(int x)
    {
        look_ahead = x;
    }

    protected boolean process_buffer(Buffer buf)
    {
        // Figure out the format of this buffer
        Caps bufcaps = buf.getCaps();

        int size = buf.getSize();
        //System.out.println("Got " + size + " byte buffer of type: " + bufcaps);
        
        Structure bcaps = bufcaps.getStructure(0);
        int rate = bcaps.getInteger("rate");
        int channels = bcaps.getInteger("channels");

        if (channels > 2)
        {
            System.err.println("Error: Too many channels " + channels);
            return false;
        }

        // Get the raw data out of the buffer
        ByteBuffer bbuf = buf.getByteBuffer();
        FloatBuffer fbuf = bbuf.asFloatBuffer();

        float[] samples = new float[size / 4];
        fbuf.get(samples);

        int num_samples = samples.length / channels;
        long dur = (long)num_samples * (1000000000/rate);

        //System.out.println("Buffer #" + buf_count + " contains " + num_samples
        //        + " samples at " + rate + " with " + channels + " channels.");

        // Do something useful with the samples here
        
        float local_max = Float.NEGATIVE_INFINITY;
        float local_min = Float.POSITIVE_INFINITY;

        for (int i = 0; i < num_samples; ++i)
        {
            // initialise because javac is stupid
            float lo_sample = 0, hi_sample = 0;

            switch (channels)
            {
                case 1:
                    lo_sample = hi_sample = samples[i];
                    break;
                case 2:
                    if (samples[2*i] < samples[2*i+1])
                    {
                        lo_sample = samples[2*i];
                        hi_sample = samples[2*i+1];
                    }
                    else
                    {
                        lo_sample = samples[2*i+1];
                        hi_sample = samples[2*i];
                    }
                    break;
            }

            if (hi_sample > local_max) local_max = hi_sample;
            if (lo_sample < local_min) local_min = lo_sample;

            if (hi_sample > max_data) max_data = hi_sample;
            if (lo_sample < min_data) min_data = lo_sample;
        }

        if (local_max > data[2*buf_count]) data[2*buf_count] = local_max;
        if (local_min < data[2*buf_count+1]) data[2*buf_count+1] = local_min;

        return true;
    }

    public int readData(int points)
    {
        // Create application sink
        AppSink as = (AppSink)ElementFactory.make("appsink", "testsink");

        // Specify our capabilities
        Caps caps = new Caps("audio/x-raw-float, width=32");
        as.setCaps(caps);

        // Create playbin
        PlayBin playbin = new PlayBin("AudioPlayer");
        playbin.setInputFile(new File(file));

        playbin.setAudioSink(as);
        playbin.setVideoSink(ElementFactory.make("fakesink", "videosink"));

        playbin.setState(State.PAUSED);
        playbin.getState();        
        length = playbin.queryDuration(Format.TIME);

        //System.out.println("Length: " + length);

        Buffer buf;

        buf = as.pullPreroll();
        final long rate = buf.getCaps().getStructure(0).getInteger("rate");
        final long samples_per_buf = buf.getSize() / 4;
        final long buf_len = new Double(((double)samples_per_buf / rate) * 1000000000).longValue();
        final long buf_tot = length / buf_len;// + 1;

        if (points > buf_tot) points = new Long(buf_tot).intValue();

        final float dt = (float)buf_tot / points;

        //System.out.format("Sampling %d points at %.2fs intevals with %d buffer (%.2fs) lookahead from a %ds length track.\n",
        //        points, (float)(dt * buf_len) / 1000000000, look_ahead, look_ahead * (float)(samples_per_buf) / rate, length / 1000000000);

        // NOTE: array is initialised to zero by jvm
        // probably should initialise to -Inf,+Inf,...
        data = new float[points * 2];

        buf_count = 0;

        long pos;

        while (buf_count < points-20)
        {
            pos = new Float(buf_count * dt * buf_len).longValue();
            //playbin.seek(pos, TimeUnit.NANOSECONDS);

            for (int i = 0; i < look_ahead; ++i)
            {
                playbin.seek(pos + i * buf_len, TimeUnit.NANOSECONDS);
                process_buffer(as.pullPreroll());
            }

            ++buf_count;

            if (threaded &&
                    index < sync.get())
            {
                term = true;
                break;
            }
        }

        if (!term)
            normaliseData();

        // Clean up
        playbin.setState(State.NULL);
        playbin.getState();

        return (term ? 0 : 1);
    }

    protected void normaliseData()
    {
        float norm = Math.max(Math.abs(max_data), Math.abs(min_data));

        for (int i = 0; i < data.length; ++i)
        {
            data[i] = Math.copySign(Math.abs(data[i]) / norm, data[i]);

            //TODO: fix this hack
            if (Math.abs(data[i]) == 1 && i > 1)
            {
                data[i] = data[i-2];
            }
        }
    }
}
