/*   This file is part of Armadillo Music Player.
*
*   Armadillo is free software: you can redistribute it and/or modify
*   it under the terms of the GNU General Public License as published by
*   the Free Software Foundation, either version 3 of the License, or
*   (at your option) any later version.
*
*   Armadillo is distributed in the hope that it will be useful,
*   but WITHOUT ANY WARRANTY; without even the implied warranty of
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*   GNU General Public License for more details.
*
*   You should have received a copy of the GNU General Public License
*   along with Armadillo.  If not, see <http://www.gnu.org/licenses/>.
*/

package armadilloUI;

import java.util.LinkedList;
import java.util.logging.Level;
import java.util.logging.Logger;

/**
 * Thread class for importing tracks in background
 * @author Kate
 */
public class ImportSongBPMThread implements Runnable {

    private final Object synchObjBPM = new Object();
    private ArmadilloDB db;
    private ArmadilloUI ui;
    private int percentage = 0;
    private volatile boolean running = true;
    private volatile boolean paused = false;
    private boolean stopped = false;
    private Thread t;
    // queue holds batches (array lists) of IDs to analyse:
    private volatile LinkedList<int[]> queue = new LinkedList<int[]>();
    private int[] batch;
    private int currentIndex;

    /**
     * Constructor: makes a new object and initialises it
     * @param name
     * @param IDs
     * @param ui
     */
    public ImportSongBPMThread(String name, int[] IDs, ArmadilloUI ui)
    {
        ArmadilloLog.LogOperation("Constructing Import song BPM thread");
        this.openDbObject();
        t = new Thread(this, name);
        this.ui = ui;
        addToQueue(IDs);
        t.start();
    }

    /**
     * Constructor: makes a new object and initialises it
     * @param name
     * @param ui
     */
    public ImportSongBPMThread(String name, ArmadilloUI ui)
    {
        ArmadilloLog.LogOperation("Constructing Import song BPM thread");
        this.openDbObject();
        t = new Thread(this, name);
        this.ui = ui;
        this.queue = db.getQueue();
        this.running = true;
        db.clearQueueTable();
        this.paused = false;
        t.setPriority(Thread.NORM_PRIORITY-2);
        t.start();
    }

    /**
     * runs the import thread
     */
    @Override
    public void run()
    {
        while (running && (queue.size() > 0) && !(paused))
        {
            calculateBPM();
        }
        this.closeDbObject();
        if (stopped)
            ui.updateProgressBar(0);
        stopped = false;
    }

    /**
     * calls the code to calculate the BPM of a track
     */
    public void calculateBPM()
    {   
        this.batch = this.queue.removeFirst();
        for (int i=0; i<batch.length; i++)
        {
            currentIndex = i;
            if (!running || paused)
            {
                //this.closeDbObject();
                return;
            }
            percentage = (((i+1)*100)/batch.length);
            ui.updateProgressBarText("Analysing "+(i+1)+" of "+batch.length);
            ui.updateProgressBar(percentage);
            String[] track = db.getTrack(batch[i]);
            if (track != null && !(db.getAnalysed(batch[i])))
            {
                String[] args = new String[1];
                args[0] = track[5];
                BPM bpm = new BPM();
                Boolean success = bpm.run(args, ui.moodDetail, false);
                if (success == true)
                {
                    Bin temp = bpm.get_bin();
                    temp.normalize();
                    // STORE BPM IN DB
                    int id = db.getTag(bpm.getBPM()+"", 5);
                    if ( id == -1)
                        id = db.insertTag(bpm.getBPM()+"", 5);
                    db.tagTrack (id, batch[i]);
                    // STORE FINGERPRINTING DATA
                    db.storeFPDoubles(batch[i], temp.gdec());
                    db.storeFPInts(batch[i], temp.ghz());
                    db.storeAverageDec(batch[i], temp.getAverageDec());
                    // testing only:
                    //db.showFingerprints(batch[i]);

                    ui.reloadNavTreeBPM();
                    ui.reloadTable(batch[i]);
                }
                db.setAnalysed("true", batch[i]);
            }
        }
        if (!paused)
        {
            percentage = 0;
            ui.updateProgressBar(percentage);
        }
    }

    /**
     * adds the passed tracks to the queue to analyse
     * @param newIDs
     */
    public void addToQueue(int[] newIDs)
    {
        queue.addLast(newIDs);
    }

    /**
     * stops the analysis process
     */
    public void stopAnalysing()
    {
        running = false;
        stopped = true;
    }

    /**
     * pauses the analysis process
     * @param visible
     * @return
     */
    public boolean pauseAnalysing(boolean visible)
    {
        if (visible)
            ui.updateProgressBarText("Paused");
        int[] remaining = new int[batch.length - currentIndex];
        for (int p=0; p<remaining.length; p++)
        {
            remaining[p] = batch[currentIndex];
            currentIndex++;
        }
        queue.addFirst(remaining);
        db.saveQueue(this.queue);
        paused = true;
        return paused;
    }

    boolean isAlive() {
        return t.isAlive();
    }

    private void closeDbObject()
    {
        this.db.closeConnection();
    }

    
    private void openDbObject()
    {
        try {
            this.db = new ArmadilloDB("org.sqlite.JDBC", "jdbc:sqlite:armadillo.db");
        } catch (Exception ex) {
            ArmadilloLog.LogError("Import song BPM thread cannot access DB - " + ex.toString());
            Logger.getLogger(ImportSongBPMThread.class.getName()).log(Level.SEVERE, null, ex);
        }
    }

    /*public void suspendBPMAnalysis()
    {
        try
        {
            synchronized (synchObjBPM) {
                synchObjBPM.wait();
            }
        }
        catch (InterruptedException iex)
        {
            System.err.println("bmpThread interrupted");
        }
    }

    public void resumeBPMAnalysis() throws InterruptedException {
        synchronized (synchObjBPM) {
            synchObjBPM.notify();
        }
    }*/
}
