/*   This file is part of Armadillo Music Player.
*
*   Armadillo is free software: you can redistribute it and/or modify
*   it under the terms of the GNU General Public License as published by
*   the Free Software Foundation, either version 3 of the License, or
*   (at your option) any later version.
*
*   Armadillo is distributed in the hope that it will be useful,
*   but WITHOUT ANY WARRANTY; without even the implied warranty of
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*   GNU General Public License for more details.
*
*   You should have received a copy of the GNU General Public License
*   along with Armadillo.  If not, see <http://www.gnu.org/licenses/>.
*/

/*
 * This class contains the main method to run the Armadillo Media Player
 */

package armadilloUI;

import java.awt.Font;
import java.util.logging.Level;
import java.util.logging.Logger;
import org.gstreamer.*;
import java.io.*;
import javax.swing.UIManager;
import javax.swing.UIManager.LookAndFeelInfo;

import armadilloUI.runcheck.RunCheck;
import java.util.ArrayList;



/**
 * The main class, this contains the entry point of the program
 * @author jonathan
 */
public class Main {

    
    static boolean initial = false;             // USED TO CHECK IF DB ALREADY EXISTS
    private static Boolean testing  = false;    // CHANGE TO FALSE WHEN NOT TESTING

    /**
     * contains boot-up sequence and checks
     * @param args
     */
    public static void main(String[] args) {
        ArmadilloLog.NewLog();
        ArmadilloLog.LogOperation("Starting Armadillo");
        UIUtilities util = new UIUtilities();
        // check for multiple instances
        if (RunCheck.isRunning())
        {
            System.out.println("Armadillo is already running!");
            new DialogUI(23).setVisible(true);
            return;
        }
        else
        {
            if (!RunCheck.setPrimaryInstance())
                RunCheck.setPrimaryInstance();
        }

        SplashScreen splash = new SplashScreen();
        splash.show();

        try {
            // CHECK IF DB HAS BEEN BUILT
            File f = new File("armadillo.db");
            if (!f.exists()) { initial = true; }
            if (testing) { System.out.println("DB exists? "+f.exists()); }

            // INITIALIZE CONNECTION WITH DB
            ArmadilloDB db = new ArmadilloDB("org.sqlite.JDBC", "jdbc:sqlite:armadillo.db");

            // IF DB HASN'T BEEN CREATED, BUILD ALL TABLES
            if (initial)
            {
                if (testing) { System.out.println("Building tables..."); }
                db.createTables();
            }
            else
                db.clearQueueTable();

            // JUST FOR NOW - TESTING DB fingerprints
            if (false)
            {
                double[] a1 = new double[49];
                int[] a2 = new int[49];
                double[] a3 = new double[49];
                int[] a4 = new int[49];
                for (int i=0; i<49; i++)
                {
                    a1[i] = 11.11;
                    a2[i] = 1;
                    a3[i] = 33.33;
                    a4[i] = 3;
                }

                db.storeFPDoubles(1, a1);
                db.storeFPInts(1, a2);
                db.storeFPDoubles(2, a3);
                db.storeFPInts(2, a4);
                db.showFingerprints(1);
                db.showFingerprints(2);
            }

            //Set up Nimbus LAF
	    //UIManager.put("nimbusBase", new java.awt.Color(0,0,0));
	    //UIManager.put("control", new java.awt.Color(0,0,0));
	    UIManager.put("text", new java.awt.Color(255,0,0));
	    UIManager.put("Table.background", new java.awt.Color(64,64,64));

            try {
                for(LookAndFeelInfo info : UIManager.getInstalledLookAndFeels()) {
                     if("Nimbus".equals(info.getName())) {
                        UIManager.setLookAndFeel(info.getClassName());
                        break;
                     }
                }
            }
            catch (Exception e){
                ArmadilloLog.LogError("Main class cannot access nimbus - "+e.toString());
                e.printStackTrace();
            }

            Font font = new Font("URW Gothic L",Font.PLAIN,12);
            java.util.Enumeration keys = UIManager.getDefaults().keys();
            while (keys.hasMoreElements()) {
              Object key = keys.nextElement();
              Object value = UIManager.get (key);
              if (value instanceof javax.swing.plaf.FontUIResource) {
                    UIManager.put (key, font);
              }
            }

            try
            {
                args = Gst.init("AudioPlayer", args);
            }
            catch (UnsatisfiedLinkError e)
            {
                ArmadilloLog.LogError("Main: gstreamer not installed - "+e.toString());
                System.err.println("Fatal error: GStreamer is not installed.");
                splash.hide();

                // pop up a dialog box
                new DialogUI(10).setVisible(true);
                return;
            }

            // INITIALIZE UI
            ArmadilloUI ui = new ArmadilloUI();
            ui.setVisible(true);
            splash.hide();

            // CHECK IF THERE ARE ANY SONGS LEFT TO ANALYSE
            boolean launchBool = true;
            try {
                String launchMood = util.config("moodOnLaunch");

                if (launchMood == null) {
                    util.config("moodOnLaunch","true");
                } else {
                    launchBool = Boolean.parseBoolean(launchMood);
                }
            } catch (Exception e) {
                ArmadilloLog.LogError("Main class cannot access config file - " +e.toString());
                e.printStackTrace();
            }
            util = null;
            if (launchBool) {
                int[] IDs = db.getTracksToAnalyse();
                if (IDs.length > 0)
                {
                    // start new thread to calculate bpms of the new tracks
                    ui.setbpmThread(new ImportSongBPMThread("bpmthread", IDs, ui));
                }
            }
        }
        catch (Exception ex) {
            ArmadilloLog.LogError("Main class cannot access DB - " +ex.toString());
            Logger.getLogger(Main.class.getName()).log(Level.SEVERE, null, ex);
        }
    }
}
